flsys = LibStub('AceAddon-3.0'):NewAddon('flsys', 'AceConsole-3.0', 'AceEvent-3.0', 'AceComm-3.0')
flsys:SetDefaultModuleState(false)

local ui -- direct pointer to flsys.ui after init
local db -- direct pointer to flsys.db after init

-- cache db settings that don't change often
local raidstocount, min_gp

-------------------------------------------------------------------------------
--- INITIALIZE/ENABLE
-------------------------------------------------------------------------------

local defaults = {
  profile = {
    raidstocount = 15,
    min_gp = 500,
    min_ep = 0,
    ep_amt = 250,
    ep_recur = 30,
    players = {},
    alts_data = {},
    dates = {},
    sortlist = {},
    next_award = 0,
    lastbroadcast = 0,
    epgplog = {},
    epgp_export = 0, --timestamp
    standings_export = 0, --timestamp
    inprogress = false,
    whispopts = true,
    debuglevel = 3,
    promptrecur = true,
    mode = 1,
    standard_ilvl = 655 -- highmaul normal
  },
}

function flsys:OnInitialize()
  local fulldb = LibStub('AceDB-3.0'):New('FlamingLootSysDB', defaults)
  flsys.log = LibStub('AceDB-3.0'):New('flsysdebuglog')
  flsys.log = flsys.log.profile
  flsys.db = fulldb.profile
  db = flsys.db
  ui = flsys.ui
end

function flsys:OnEnable()
  flsys:Info('flsys enabled')
  flsys:InitializeComm()
  flsys:RegisterChatCommand('fss', function() FLMain:Show() end)
  flsys:RegisterChatCommand('fs', 'ProcessChatCmd')
  flsys:GetConstants()
  ui:Initialize()
  if db.inprogress then -- this check needs to happen later
    flsys:InRaidCheck(true)
  end
  flsys:Print('Flaming loot masterlooter addon loaded.  Type /fs for a list of commands.')
  flsys:TrimDebug()
  flsys:CheckForGameUpdate()
end

function flsys:GetConstants()
  --load db constants into local memory
  raidstocount = db.raidstocount
  min_gp = db.min_gp
end


function flsys:CheckForGameUpdate()
  -- prompt to change default ilvl if game version has changed
  if not db.standard_ilvl then
      db.standard_ilvl = defaults.profile.standard_ilvl
  end

  if defaults.profile.standard_ilvl > db.standard_ilvl then
    local old_vers = db.game_vers or 0
    local current = select(1, GetBuildInfo())
    current = string.gsub(current, '%.', '')
    current = tonumber(current)

    if current >= old_vers then 
      StaticPopup_Show('CHANGE_ILVL')
    end
  end
end

-------------------------------------------------------------------------------
--- GP
-------------------------------------------------------------------------------

function flsys:OnLootAwarded(name, link, id, amount)
  if db.players[name] then
    flsys:AddGP(name, amount, link, id)
  else
    flsys:Error("OnLootAwarded but player is not in db. Player: "..name)
  end
end


function flsys:AddGP(name, amount, itemname, id, raid_i)
  flsys:Info('Adding GP: '..name.." "..amount.." "..itemname)
  --scale
  local r_i = raid_i or 1
  amount = flsys:Round(amount*(raidstocount + 1 - r_i)/raidstocount, 0)
  --add to total & raid
  db.players[name].gp_data[r_i] = math.max(db.players[name].gp_data[r_i] + amount, 0)
  db.players[name].gp_total = math.max(db.players[name].gp_total + amount, 0)
  --don't add log if this is retroactive; let those funcs handle it
  if not raid_i then --only for auto (not log edits)
    flsys:Log(amount, 'GP', name, itemname, id)
  end
  flsys:UpdatePR()
end


--administration
function flsys:ScaleGP(oldstd, newstd)
  local div = 2^((newstd-oldstd)/26)
  for name, nt in pairs(db.players) do
    nt.gp_total = flsys:Round(nt.gp_total/div)
    for i, g in ipairs(nt.gp_data) do
      nt.gp_data[i] = flsys:Round(g/div)
    end
  end
  flsys:Log(0, 'decay', 'all', 'GP scaled: '..oldstd..' to '..newstd, 0)
  flsys:UpdatePR()
end

function flsys:ResetGP()
  for name, nt in pairs(db.players) do
    nt.gp_total = 0
    for i, g in ipairs(nt.gp_data) do
      nt.gp_data[i] = 0
    end
  end
  flsys:Log(0, 'RESET', 'all', 'Reset all GP')
  flsys:UpdatePR()
  flsys:Print('Reset all GP.')
end

-------------------------------------------------------------------------------
--- EP
-------------------------------------------------------------------------------

function flsys:AddEP(name, amount, reason, mass, raid_i)
  -- if alt, add ep to main's record
  if db.players[name].alt then
    name = db.alts_data[name]
  end
  --scale based on raid index (assume current)
  local r_i = raid_i or 1
  amount = flsys:Round(amount*(raidstocount + 1 - r_i)/raidstocount, 0) --scale
  --update raid and total
  db.players[name].ep_data[r_i] = math.max(db.players[name].ep_data[r_i] + amount, 0)
  db.players[name].ep_total = math.max(db.players[name].ep_total + amount, 0)
  --update pr, add log entry
  if not mass then
    flsys:Info('Adding single EP: '..name..' '..amount..' raid i: '..(raid_i or 'none'))
    flsys:UpdatePR()
    if not raid_i then -- only for auto awards (not log edits)
      flsys:Log(amount, 'EP', name, reason)
    end
  end
end

function flsys:MassEP(amount, reason)
  if db.inprogress then
    local namelist = ''
    for name,_ in pairs(db.players) do
      if db.players[name].inraid then
        flsys:AddEP(name, amount, reason, true)
        namelist = namelist..name..', '
      end
    end
    flsys:Log(amount, 'EP', string.sub(namelist, 1, -3), reason)
    flsys:Info("Mass EP of "..amount.." awarded")
    flsys:UpdatePR()
    if db.mode == 1 then
      flsys:Print(amount..'EP awarded to raid.')
    end
  else
    flsys:Warning("Tried to add mass ep without raid in progress")
    flsys:Print('There is no active raid.')
  end
end
-------------------------------------------------------------------------------
--- PR & SORTING
-------------------------------------------------------------------------------

function flsys:UpdatePR()
  --- gets each players PR, calls sort, sends standings if in raid
  --BY: StartRaid, EndRaid, AddEP, AddGP, MassEP, RefreshSortList
  --CALLS: UpdateMain

  flsys:Info('Updating PR for all players')
  for name, i in pairs(db.players) do
    i.current_pr = flsys:GetPR(name)
  end

  flsys:SortRoster()

  if db.inprogress then
    flsys:SendStandings('raid')
  end
end

function flsys:GetPR(name)
  --- Returns a player's PR, 0 if N/A

  local player = db.players[name]

  -- pr does not apply to alts, or to subs in ongoing mode
  if (db.mode == 1 and (player.alt and player.temp)) then
    return 0
  end

  if db.mode == 2.2 then min_gp = .5 end
  local ep = player.ep_total
  local gp = math.max(player.gp_total, min_gp)
  return flsys:Round(ep/gp, 2)
end

function flsys:GetPlayerDetails(name, short)
  local player = db.players[name]
  local mode = db.mode

  local status
  if mode == 1 then
    if player.temp then
      status = 'sub'
    elseif player.alt then
      status = 'alt'
    end
  end
  if short then
    if flsys:IsBelowMinEP(name) then
      status = 'below min EP'
    end
    return status
  end
  
  if status then
    status = flsys:Cap(status)
  else
    status = 'Team member'
  end

  local reason
  if mode == 1 then
    if status == 'Alt' then
      reason = 'Alts have lower priority than mains on loot, so they get an automatic PR of 0.'..
      '\nThey receive EP which counts toward the PR of their main.'..
      '\nGP received only counts for the current raid.'
    elseif status == 'Sub' then
      reason = 'Subs do not have PR. If they roll need against a team member'..
      '\nit is determined by a random roll. Once they have received any gear,'..
      '\nthey have lowest priority for the rest of the raid.'
    else
      if flsys:IsBelowMinEP(name) then
        reason = 'This player is below the minimum EP ('..db.min_ep..'), so they effectively'..
        '\nhave 0 PR, regardless of EP, until they reach the minimum.'..
        '\nThey can still win gear no one else needs, and can roll against subs,'..
        '\nbut will not win main-spec loot against a team member.'
      elseif player.gp_total < db.min_gp then
        reason = 'A minimum GP is used to calculate PR.\n'..
          player.ep_total..' (EP)/'..db.min_gp..' (min GP) = '..player.current_pr..' (PR)'
      else
        reason = player.ep_total..'(EP)/'..player.gp_total..'(GP) = '..player.current_pr..' (PR)'
      end
    end
  elseif mode == 2.1 then
    if (player.current_pr == player.ep_total / player.gp_total) then
      reason = player.ep_total..'(EP)/'..player.gp_total..'(GP) = '..player.current_pr..' (PR)'..
        '\n(In this mode, a large EP award is given at the start'..
        '\nof raid to make the PR numbers more spread out.'
    end
  else
    reason = 'This player has received '..player.gp_total..' item(s) this raid.'
  end

  return status, reason or ''
end


function flsys:DecayRaid()
  for name,_ in pairs(db.players) do
    db.players[name].ep_total = 0
    db.players[name].gp_total = 0
    local start = #db.players[name].ep_data
    if start < raidstocount then
      start = start + 1
    end
    for i = start, 2, -1 do
      db.players[name].ep_data[i] = math.floor((db.players[name].ep_data[i-1] or 0)*(raidstocount-i + 1)/(raidstocount-i + 2))
      db.players[name].gp_data[i] = math.floor((db.players[name].gp_data[i-1] or 0)*(raidstocount-i + 1)/(raidstocount-i + 2))
      db.players[name].ep_total = db.players[name].ep_total + db.players[name].ep_data[i]
      db.players[name].gp_total = db.players[name].gp_total + db.players[name].gp_data[i]
    end
    db.players[name].ep_data[1] = 0
    db.players[name].gp_data[1] = 0
  end
  flsys:Print('EP and GP decayed.')
  flsys:Info('Decayed')
  flsys:Log(0, 'decay', 'all', 'Decayed EP and GP')
  flsys:UpdatePR()
end

-------------------------------------------------------------------------------
--- MAJOR SETTINGS CHANGES
-------------------------------------------------------------------------------

function flsys:ModeChanged(mode)
  if mode == 1 then
    if not db.players[GetUnitName('player')] then
      flsys:NewPlayer(GetUnitName('player'))
    end
  else
    flsys:ResetData(true)
  end
end

function flsys:ResetData(removeplayers)
  wipe(db.dates)
  wipe(db.epgplog)
  if removeplayers then
    wipe(db.players)
    wipe(db.alts_data)
    flsys:Print('Removed all players and their data from the system.')
    flsys:Log(0, 'RESET', 'all', 'Reset & removed all players')
  else
    for name, ntable in pairs(db.players) do
      flsys:NewPlayer(name)
    end
    flsys:Print('Reset data for all players.')
    flsys:Log(0, 'RESET', 'all', 'Reset all EP and GP')
  end
  flsys:RefreshSortList()
end

function flsys:ImportEntered(imptext, fromplayer)
  if db.inprogress then
    flsys:Print('You cannot import or receive data while you have a raid in progress.')
    return
  end
  if imptext and imptext ~= '' then
    local printmsg, sendplayer, sendplayererror
    for psection in string.gmatch(imptext, '@([^@]+)') do
      local player = string.match(psection, '^([^=]+)')
      local eppos = string.find (psection, 'ep:')
      local balpos = string.find (psection, 'bal:') or -1
      if not player or not eppos or not string.find(psection, 'gp:') then
        printmsg = 'Import text not formatted correctly. Make sure you paste the entire text exactly as it is in the spreadsheet.  If you\'re seeing this error after receiving data from another player, please go to curse.com/addons/wow/flaming-ml and report a bug.'
        sendplayererror = 'could not import your data correctly.  This is probably an addon bug.  Please go to curse.com/addons/wow/flaming-ml and report it.'
      else
        FLImportData:Hide()
        flsys:NewPlayer(player)
        local currpart = string.sub(psection, 1, eppos)
        local pos = 1
        for amt in string.gmatch(currpart, '(%d+)') do --filling in gp data, updating total
          amt = tonumber(amt)
          db.players[player].gp_data[pos] = amt
          db.players[player].gp_total = db.players[player].gp_total +amt
          pos = pos + 1
        end
        currpart = string.sub(psection, eppos, balpos)
        pos = 1
        for amt in string.gmatch(currpart, '(%d+)') do --filling in epdata, updating total
          amt = tonumber(amt)
          db.players[player].ep_data[pos] = amt
          db.players[player].ep_total = db.players[player].ep_total +amt
          pos = pos + 1
        end
        local bal = string.match(psection, 'bal:(%-?%d+)')
        if bal then
          db.players[player].balance = tonumber(bal)
        end
        flsys:Log(0, 'import', player, 'Imported data for '..player)
        flsys:Print('Imported data for '..player..'.')
        flsys:UpdatePR()
        sendplayer = 'has imported your data!'
      end
    end
    flsys:Print(printmsg)
    if fromplayer then
      if sendplayererror then
        SendAddonMessage('flamingsys', 'datareply:'..sendplayererror, 'WHISPER', fromplayer)
      elseif sendplayer then
        SendAddonMessage('flamingsys', 'datareply:'..sendplayer, 'WHISPER', fromplayer)
      end
    end
  else
    flsys:Print('You entered a blank import.  Copy the text from your data backup spreadsheet.  If you\'re seeing this error after trying to receive data from another player, confirm that they have data stored - if so, please go to curse.com/addons/wow/flaming-ml and report a bug.')
    if fromplayer then
      SendAddonMessage('flamingsys', 'datareply: did not receive your data - either you have no data stored or there was an addon bug.  Please go to curse.com/addons/wow/flaming-ml and report it.', 'WHISPER', fromplayer)
    end
  end
end

function flsys:SendData(name)
  flsys:SendCommMessage('flamingsysdata', 'data:'..flsys:ExportToString(), 'WHISPER', name)
  flsys:Print('Sending data to '..name..'.  If they accept or decline the data you\'ll see a printout.  No further message means they don\'t have the addon turned on or they\'re using an outdated version.')
end

-------------------------------------------------------------------------------
--- ROSTER
-------------------------------------------------------------------------------

local numraidmembers

--- TODO EVENTS notes
-- number of players changed - updatesortlist, updateraidroster if in raid,
--   checkrows, renderrows
-- PR updated - updatesortlist, renderrows
-- raid started or ended - updateraidroster, updatesortlist, checkrows

-- only relevant in a raid!
function flsys:UpdateRaidRoster()
  -- compiles a raid roster by setting inraid and adding new players as needed

  for name, _ in pairs(db.players) do
    db.players[name].inraid=nil
  end

  numraidmembers = GetNumGroupMembers()
  flsys:Info("UpdateRaidRoster(), "..numraidmembers.." members")
  local name
  for i = 1, numraidmembers do
    name = select(1, GetRaidRosterInfo(i))
    if db.players[name] then
      db.players[name].inraid = true
    elseif db.alts_data[name] and db.players[db.alts_data[name]] then
      flsys:NewPlayer(name, true, false, true)
    else
      flsys:NewPlayer(name, true, true)
    end
  end
  flsys:RefreshSortList()
end

-- call when db.players is altered (player added/removed, including temps/alts)
function flsys:RefreshSortList()
  --- Wipes & recreates sortlist containing all players
  wipe(db.sortlist)
  for name, t in pairs(db.players) do
    if t then
      table.insert(db.sortlist, name)
    end
  end
  flsys:SortRoster()
  flsys.ui:CheckRows()
  flsys.ui:InitializeRoster()
end


function flsys:SortRoster()
  -- sorts by: inraid > temp > PR

  flsys:Debug('Sorting sortlist')

  table.sort(db.sortlist, function(name1, name2)
    -- return true if name1 should be first, else false
    if db.players[name1].inraid ~= db.players[name2].inraid then
      return db.players[name1].inraid
    elseif db.players[name1].temp ~= db.players[name2].temp then
      return db.players[name2].temp
    else
      return flsys:SortByPR(name1, name2)
    end
  end)

  flsys.ui:RenderStandings()
end

function flsys:IsBelowMinEP(name)
  return db.players[name].ep_total < db.min_ep
end

function flsys:SortByPR(name1, name2)
  -- returns true if name1 should be ranked higher, else false
  -- accounts for min EP. if tied, player with less GP is ranked first.
  if flsys:IsBelowMinEP(name1) ~= flsys:IsBelowMinEP(name2) then
    return flsys:IsBelowMinEP(name2)
  elseif db.players[name1].current_pr ~= db.players[name2].current_pr then
    return db.players[name1].current_pr > db.players[name2].current_pr
  else
    return db.players[name1].gp_total < db.players[name2].gp_total
  end
end


function flsys:GROUP_ROSTER_UPDATE()
  if not IsInRaid() then
    flsys:Info("GROUP_ROSTER_UPDATE - user no longer in raid")
    StaticPopup_Show("ENDRAID")
  elseif GetNumGroupMembers() ~= numraidmembers then
    flsys:Debug("GROUP_ROSTER_UPDATE - # raid members changed")
    flsys:UpdateRaidRoster()
    fld:CreateRoster()
  end
end

-------------------------------------------------------------------------------
--- PLAYER DATABASE
-------------------------------------------------------------------------------

function flsys:NewPlayer(name, inraid, temp, alt, ep, gp)
  --- Adds a player to the database
  -- calls: RefreshSortList, ~UpdateRaid
  db.players[name] = {
    inraid = inraid,
    temp = temp or false,
    alt = alt,
    balance = 0,
    ep_data = {[1] = 0},
    gp_data = {[1] = 0}
  }
  if ep and gp then
    local ep = flsys:Round(ep*(db.raidstocount-1)/(db.raidstocount))
    local gp = flsys:Round(gp*(db.raidstocount-1)/(db.raidstocount))
    db.players[name].ep_total = ep
    db.players[name].gp_total = gp
    db.players[name].ep_data[2] = ep
    db.players[name].gp_data[2] = gp
  else
    db.players[name].ep_total = 0
    db.players[name].gp_total = 0
  end
  db.players[name].current_pr = flsys:GetPR(name)
  flsys:Info('Added '..name..' to team with PR of '..db.players[name].current_pr)
  if db.inprogress then
    flsys:UpdateRaidRoster()
  else
    flsys:RefreshSortList()
  end
end


function flsys:AddAlt(info, input)
  local aremove, alt, main = string.match(input, "(-*)([^%d%s%p]+)=*([^%d%s%p]*)")
  if alt then alt = flsys:Cap(alt) end
  if main then main = flsys:Cap(main) end
  if aremove and db.alts_data[alt] then
    db.alts_data[alt] = nil
    flsys:Print(alt.." removed from alts list.")
  elseif alt and main then
    if not db.players[main] then
      self:Print(main.." is not in the database.  Add them before adding their alts.")
    else
      db.alts_data[alt] = main
      flsys:Print(alt.." added to alts list.")
    end
  else
    flsys:Print("Invalid entry. Make sure to format 'altname=mainname' without spaces")
  end
  alts_data = db.alts_data
end


function flsys:RemovePlayer(name)
  FLRosterFrame:Hide()
  if db.inprogress then
    self:Print("Cannot remove players during a raid.")
  else
    flsys:Info("Removing "..name.." from database.")
    db.players[name] = nil
    if db.mode==1 then self:Print(name.." removed from database.") end
    flsys:RefreshSortList()
  end
end

-------------------------------------------------------------------------------
--- RAID
-------------------------------------------------------------------------------

function flsys:StartRaid(raidname)
  local mlid = select(3, GetLootMethod())
  if flsys.db.inprogress then
    flsys:Print('You already have a raid in progress.')
  elseif not IsInRaid() then
    flsys:Print('You are not in a raid.')
  elseif not mlid then
    flsys:Print('You must be the masterlooter to start a raid.')
  elseif select(1, GetRaidRosterInfo(mlid)) ~= GetUnitName('player') then
    flsys:Print('You must be the masterlooter to start a raid.')
  else --prereqs met; start the raid
    flsys:Info('Starting raid *****')
    fld:Enable()
    flsys:RegisterEvent('GROUP_ROSTER_UPDATE')
    flsys.db.inprogress = raidname or 'Raid'
    flsys.ui:MainButtons()
    table.insert(flsys.db.dates, 1, date('%m/%d %H:%M'))
    flsys:Log(0, 'start', 'all', flsys.db.inprogress..' started', 0)  --FIX
    flsys:UpdateRaidRoster()
    fld:StartRaid()

    flsys:SendVersion()

    if flsys.db.mode == 1 then
      if flsys.db.promptrecur then
        StaticPopup_Show('START_RECUR')
      end
    elseif flsys.db.mode == 2.1 then
      flsys:MassEP(4000, 'single, GP')
    elseif flsys.db.mode == 2.2 then
      flsys:MassEP(3, 'single, count')
    end
  end
end

-- Frame used as timer
local recur_frame = CreateFrame('Frame', 'flsys_TimerFrame')
recur_frame:SetScript('OnUpdate', function(self, elapsed) flsys:CheckTime(self, elapsed) end)
recur_frame:Hide()

function flsys:ResumeRaid()
  flsys:Info('ResumeRaid()')
  flsys:UpdateRaidRoster()
  if flsys.db.next_award > 0 then
    flsys:Info('Recurring award resumed')
    flsys:Print('Recurring EP award resumed.')
    recur_frame:Show()
  end
  flsys:RegisterEvent('GROUP_ROSTER_UPDATE')
  fld:Enable()
  fld:StartRaid()
end

function flsys:EndRaid()
  if flsys.db.inprogress then
    flsys:Info('EndRaid() *****')
    flsys:Log(0, 'end', 'all', 'Raid ended') --TODO customize which raid
    flsys:Print('Raid ended.')
    flsys.db.dates[1] = flsys.db.dates[1]..' - '..date('%m/%d %H:%M')..' '..flsys.db.inprogress
    table.insert(flsys.db.dates, 1, '')
    flsys.db.inprogress = false
    local player_removed = false
    for name, ntable in pairs(flsys.db.players) do
      if flsys.db.players[name].temp or flsys.db.players[name].alt then
        flsys:Debug('Removing', name, 'from system (alt or sub)')
        flsys:RemovePlayer(name)
        player_removed = true
      else
        flsys.db.players[name].inraid = nil
      end
    end
    flsys:EndRecur()
    flsys:UnregisterEvent('GROUP_ROSTER_UPDATE')
    flsys:UnregisterEvent('COMBAT_LOG_EVENT_UNFILTERED')
    flsys:UnregisterEvent('TRADE_ACCEPT_UPDATE')
    flsys:Info("Disabling fld")
    fld:Disable()
    flsys.ui:MainButtons()
    if flsys.db.mode == 1 then
      flsys:DecayRaid()
    end
    if player_removed then
      flsys:RefreshSortList()
    end
  else
    flsys:Print('You do not have a raid in progress.')
    flsys:Warning('EndRaid called without raid in progress')
  end
end


function flsys:StartRecur(minutes) --TODO should do this check before the popup
  if flsys.db.next_award > 0 then
    flsys:Print('You already have a recurring award in progress.  /fs endrecur to end it.')
  elseif minutes >= 0 and minutes < 60 and flsys.db.inprogress then
    local start = GetTime() + minutes*60
    flsys.db.next_award = start
    flsys:Print('Recurring award starts in '..minutes..'min. '..flsys.db.ep_amt..' EP will be awarded every '..flsys.db.ep_recur..' min.')
    flsys:Debug('Recurring award starts in '..minutes..'min. '..flsys.db.ep_amt..' EP will be awarded every '..flsys.db.ep_recur..' min.')
    recur_frame:Show()
  elseif not flsys.db.sortlist[1] then --TODO why this check and not inprogress?
    flsys:Print('Recurring awards can only take place during an active raid.  /fs startraid to begin one.')
  else
    flsys:Print('Invalid entry.  Type /fs startrecur to bring up the window again.  Enter a number between 0 and 60.')
  end
end

function flsys:EndRecur()
  flsys.db.next_award = 0
  flsys:Debug('Ending recurring')
  recur_frame:Hide()
end

local interval = 0
local currtime
function flsys:CheckTime(self, elapsed)
  --- adapted from the frame/func in epgp_recurring
  -- called by: recur_frame.OnUpdate
  currtime = GetTime()
  interval = interval + elapsed
  if interval > 30 then
    interval = 0
    if currtime > flsys.db.next_award then
      flsys:MassEP(flsys.db.ep_amt, flsys.db.inprogress)
      flsys.db.next_award = flsys.db.next_award + flsys.db.ep_recur*60
    end
  end
end

-- After login, check if player is in raid. Retry because sometimes OnEnabled
-- fires before raid group can be detected

local delay_fra = CreateFrame('Frame', 'flsys_DelayTimerFrame')
local raid_check_timer = 0

function flsys:InRaidCheck(first_try)
  -- after login, if flsys.db.inprogress is true
  if IsInRaid() then
    flsys:Info('Prompting to resume raid')
    StaticPopup_Show('RESUME_RAID')
  elseif first_try then
    delay_fra:SetScript('OnUpdate', function(self, elapsed) flsys:DelayRaidCheck(self, elapsed) end)
    delay_fra:Show()
  else
    flsys:Warning('Second InRaidCheck, ending raid automatically')
    flsys:EndRaid()
  end
end

function flsys:DelayRaidCheck(self, elapsed)
  raid_check_timer = raid_check_timer + elapsed
  if raid_check_timer > 2 then
    flsys:InRaidCheck(false)
    delay_fra:Hide()
  end
end

-------------------------------------------------------------------------------
--- LOG (IN-GAME)
-------------------------------------------------------------------------------

function flsys:Log(amount, logtype, player, reason, num)
  local timestamp = time()
  table.insert(flsys.db.epgplog, 1, {timestamp, amount, logtype, player, reason, num})
  flsys.ui:LogChanged()
end


function flsys:AlterLog(log_i, raid_i, alter_type)
  --- takes form input and updates the log, ep & gp accordingly
  -- uses Undo, AddEP and AddGP to handle epgp, modifies the log itself
  -- @param alter_type - "edit" or "add"

  local amount = FLLogItemAmount:GetNumber()
  local log_type = FLLogItemType:GetText()
  local player = UIDropDownMenu_GetSelectedName(FLRosterDropDown)
  local reason = FLLogItemReason:GetText()

  flsys:Info('AlterLog')

  if not (amount and player and reason) then
    flsys:Warning('Didn\'t fill out all the fields for altering log')
    flsys:Print("Please fill out all the fields.")
    return
  end

  if alter_type == "edit" then
    local old_item = flsys.db.epgplog[log_i]
    if amount == old_item[2] and player == old_item[4] then
      -- don't need to touch playerflsys.db, just update reason in log
      flsys.db.epgplog[log_i][5] = reason
      flsys:Info("No changes made to ep or gp. Reason updated: "..reason)
      flsys:Print("No changes made to ep or gp.  Log modified")
      return
    else
      flsys:Undo(log_i, raid_i, false)
    end
  end

  -- add to new player
  if log_type == "EP" then
    flsys:AddEP(player, amount, reason, false, raid_i)
  elseif log_type == "GP" then
    flsys:AddGP(player, amount, reason, 0, raid_i)
  end
  if alter_type == "edit" then -- update log entry
    flsys.db.epgplog[log_i][2] = amount
    flsys.db.epgplog[log_i][4] = player
    flsys.db.epgplog[log_i][5] = reason
    flsys:Info('Log entry edited.')
    flsys:Print("Change saved.")
  else -- add log entry
    -- insert before log_i so it is more recent than the raid start
    if raid_i == 1 then --chrono order for current raid
      log_i = 1
    end
    table.insert(flsys.db.epgplog, log_i, {time(), amount, log_type, player, reason})
    flsys:Info('Log entry added')
    flsys:Print("Award added.")
  end
  FLLogItem:Hide()
  flsys.ui:UpdateLogFrame()
end

function flsys:Undo(log_i, raid_i, remove)
  --- Undo an existing EP or GP award.
  -- @remove indicates whether to delete the log entry & update everything
  --    (true when deleting, false when part of an edit)

  local _, amount, logtype, player, _, _ = unpack(flsys.db.epgplog[log_i])
  amount = flsys:Round(amount*(flsys.db.raidstocount + 1 - raid_i)/flsys.db.raidstocount, 0) --scale

  flsys:Info('Undo called for '..player..' '..logtype..' '..amount)
  if logtype=="EP" then
    for name in string.gmatch(player, "([^, ]+)") do
      if flsys.db.players[player] then
        flsys.db.players[name].ep_total = flsys.db.players[name].ep_total - amount
        flsys.db.players[name].ep_data[raid_i] = flsys.db.players[name].ep_data[raid_i] - amount
      else
        local warning = "Could not undo EP award for "..player.."; they were not found in the database"
        flsys:Warning(warning)
        flsys:Print(warning)
      end
    end
  elseif logtype=="GP" then
    if flsys.db.players[player] then
      flsys.db.players[player].gp_total = flsys.db.players[player].gp_total - amount
      flsys.db.players[player].gp_data[raid_i] = flsys.db.players[player].gp_data[raid_i] - amount
    else
      local warning = "Could not undo GP for "..player.."; they are no longer in the database."
      flsys:Warning(warning)
      flsys:Print(warning)
      return
    end
  end

  if remove then
    flsys:Debug('Removed log entry')
    table.remove(flsys.db.epgplog, log_i)
    flsys.ui:UpdateLogFrame() --refresh it
    flsys:UpdatePR() --refresh standings
  end
end

function flsys:ClearLog(logtype)
  FLLog:Hide()
  if not logtype or logtype=="epgp" then
    wipe(flsys.db.epgplog)
    flsys:Print("EPGP log cleared")
  end
  -- don't need to call UpdateLog; it will get called when log is shown
end


function flsys:LogItemtoString(logitem)
  local displaystring = "  "
  local logdate, amount, logtype, players, reason, num = unpack(logitem)
  if logtype=="start" or logtype=="import" or logtype=="RESET" then
    displaystring = "|cffCCFFFF"..date("%m/%d/%y %I:%M%p", logdate).." "..reason.." "
    displaystring = displaystring.."|r"
  elseif logtype=="end" or logtype=="decay" then
    displaystring = displaystring..date("%I:%M%p", logdate).."  "..reason
  else
    displaystring = displaystring..date("%I:%M%p", logdate).."  "
    if logtype=="EP" then
      if amount>0 then amount= "+"..amount end
      displaystring = displaystring..amount.." EP "..players.." ("..reason..")"
    elseif logtype=="GP" then
      displaystring = displaystring..reason.." "..players.." ("..amount.." GP)"
    end
  end
  return displaystring
end

function flsys:ExportToString() --TODO not during combat?
  -- Provides a string that could be used to import all player epgp data
  local exportstr =""
  for name, nt in pairs(flsys.db.players) do
    exportstr = exportstr.."@"..name.."=gp:"
    for i, amt in ipairs(flsys.db.players[name].gp_data) do
      exportstr = exportstr..amt..","
    end
    exportstr = exportstr.."ep:"
    for i, amt in ipairs(flsys.db.players[name].ep_data) do
      exportstr = exportstr..amt..","
    end
  end
  return exportstr
end


StaticPopupDialogs["DELETE_LOG_ITEM"] = {
  preferredIndex = STATICPOPUPS_NUMDIALOGS,
  text = "Are you sure you want to delete this award?  It will be removed from the log and the corresponding EP or GP will be removed from the player's total.",
  button1 = ACCEPT,
  button2 = CANCEL,
  OnAccept = function(self)
    flsys:Warning("ALTER LOG", 'delete on accept')
    FLLogItem:Hide()
    flsys:Undo(self.log_i, self.raid_i, true)
  end,
}


StaticPopupDialogs["EXPORT_COMPLETE"] = {
  preferredIndex = STATICPOPUPS_NUMDIALOGS,
  text = "Did you successfully back up your data?  If so the %s log will be cleared.",
  button1 = "Yes",
  button2 = "No",
  timeout = 0,
  whileDead = 1,
  OnAccept = function(self)
    flsys:ClearLog(self.etype)
  end,
}

-------------------------------------------------------------------------------
--- COMM
-------------------------------------------------------------------------------

function flsys:InitializeComm()
  self:RegisterComm("flamingsysdata") -- for syncing data between mls
  self:RegisterEvent("CHAT_MSG_ADDON")
  RegisterAddonMessagePrefix("flamingsys")
end

function flsys:SendStandings(sender, uptodate)
  if uptodate then
    SendAddonMessage("flamingloot", "uptodate:", "WHISPER", sender)
  else
    sstr =""
    for i, name in pairs(db.sortlist) do
      sstr = sstr..name..": "
      if db.mode ~= 2.2 then
        sstr = sstr..db.players[name].current_pr
      else
        sstr = sstr..db.players[name].gp_total.."\n"
      end
      local status = flsys:GetPlayerDetails(name, true)
      if status then
        sstr = sstr..'  ('..status..')'
      end
      sstr = sstr.."\n"
    end
    if sender=="raid" then
      SendAddonMessage("flamingloot", "standings:"..sstr, "RAID")
      db.lastbroadcast = time()
    else
      SendAddonMessage("flamingloot", "standings:"..sstr, "WHISPER", sender)
    end
  end
end

function flsys:SendVersion()
  local vers = GetAddOnMetadata("flamingloot", "Version")
  local one, two = string.match(vers, "(%d).*(%d)")
  if one and two then
    SendAddonMessage("flamingloot", "version:"..one..two, "RAID")
    flsys:Info("Sent check to raid: version: "..one..two)
  end
end

function flsys:CHAT_MSG_ADDON(event, prefix, message, channel, sender)
  if prefix=="flamingsys" then
    local command, value = string.match(message, "(%a+):*(.*)")
    if command=="datareply" then
      flsys:Print(sender.." "..value)
    elseif command=="standings" then
      flsys:SendStandings(sender, (tonumber(value))<db.lastbroadcast)
    end
  end
end

function flsys:OnCommReceived(prefix, message, channel, sender)
  if prefix == "flamingsysdata" then
    local command, value = string.match(message, "(%a+):*(.*)")
    if command == "data" then
      local dialog = StaticPopup_Show("RECEIVE_DATA", sender)
      if dialog then
        dialog.text = value
        dialog.player = sender
      end
    end
  end
end

-------------------------------------------------------------------------------
--- CHAT COMMANDS
-------------------------------------------------------------------------------

local cmdListOutput

function cmdList()
  flsys:Debug('/fs showing commands')
  if not cmdListOutput then
    local color = '|cffffff78'
    local reset = '|r'
    local delim = ':  '
    local desc = {
      fss = 'View the main standings frame.',
      loot = '(During a raid only) View the most recent mob\'s drops and all the rolls made on them. Only available after the rolling period ends and only for the most recent mob.',
      opt = 'View the options window. Also contains a help panel.',
      gp = 'View a printout of GP values for an ilvl. (Ex: /fs gp 670)',
      test = 'Simulate loot drops to practice/test the rolling and awarding system. You must start a raid first (be the ML in a raid group, hit Start on the standings window). Doesn\'t require other people in your group, but will work with them.'
    }
    cmdListOutput = {
      color, '/fss', reset, delim, desc.fss, '\n',
      color, '/loot', reset, delim, desc.loot, '\n',
      color, '/fs opt', reset, delim, desc.opt, '\n',
      color, '/fs gp [ilvl]', reset, delim, desc.gp, '\n',
      color, '/fs test', reset, delim, desc.test,
      '\n\nFor more information, see the help tab of the options panel.'
    }
    cmdListOutput = table.concat(cmdListOutput, '')
  end
  flsys:Print(cmdListOutput)
end

local function cmdShowOptions()
  flsys:Info('/fs opt entered, showing options window')
  if not flopt:IsEnabled() then
    fld:Enable()
    flopt:Enable()
  end
  InterfaceOptionsFrame_OpenToCategory("flaming loot ml")
end

local function cmdPrintGPValues(ilvl)
  -- within 100 of standard_ilvl
  if (-100 < ilvl - db.standard_ilvl) and (ilvl - db.standard_ilvl < 100) then
    flsys:Debug('/fs gp printing gp values for '..ilvl)
    local mult = flsys:Round((2^((ilvl-db.standard_ilvl)/26)), 3)
    local output = {
      'GP values for ilvl ', ilvl, ':', 
      '\n(Given current standard ilvl of '..db.standard_ilvl..')',
      '\n2H weapons, hunter ranged: ', (2000 * mult),
      "\nCaster/healer main hand, tank shield: ", (1500 * mult),
      "\nDual wield weapon, chest, helm, legs, trinket: ", (1000 * mult),
      "\nShoulder, belt, boots, gloves: ", (750 * mult),
      "\nOff hand, tank mainhand, neck, ring, cloak, bracers: ", (500 * mult)
    }
    flsys:Print(table.concat(output, ''))
  else
    flsys:Info('/fs gp invalid ilvl '..ilvl)
    flsys:Print("Not a valid item level. Try again.")
  end
end

function flsys:ProcessChatCmd(input)
  if not input or flsys:IsEmptyString(input) then
    cmdList()
  elseif input == 'opt' then
    cmdShowOptions()
  elseif input == 'test' then
    if not fld:IsEnabled() then
      fld:Enable()
    end
    fld:test()
  else
    local ilvl = tonumber(string.match(input, "gp (%d+)"))
    if ilvl then
      cmdPrintGPValues(ilvl)
    else
      flsys:Print("Not a valid command. Type /fs to view a list of commands")
    end
  end
end


-------------------------------------------------------------------------------
--- POPUP CALLS TODO are these necessary, could just do _show directly?
-------------------------------------------------------------------------------
function flsys:AsktoSetupRecur()
  if db.next_award == 0 then
    StaticPopup_Show('START_RECUR')
  end
end

function flsys:CheckEndRecur()
  StaticPopup_Show('ENDRECUR')
end

function flsys:DecayPressed()
  StaticPopup_Show('DECAY')
end

function flsys:ResetDataPressed() --inprogress check done
  StaticPopup_Show('DATARESET')
end

function flsys:ResetLogPressed()
  StaticPopup_Show('LOGRESET')
end

function flsys:EndPressed()
  StaticPopup_Show('ENDRAID')
end

function flsys:RemovePlayerPressed(player)
  if player then
    local dialog = StaticPopup_Show('REMOVEPLAYER', player)
    if dialog then
      dialog.player = player
      dialog.enterClicksFirstButton = true
    end
  end
end

function flsys:ResetGPPressed()
  StaticPopup_Show('RESET_GP')
end

-------------------------------------------------------------------------------
--- POPUP DIALOGS
-------------------------------------------------------------------------------
StaticPopupDialogs['MASSEP'] = {
  text = 'Mass EP',
  button1 = 'Single',
  button2 = CANCEL,
  button3 = 'Recurring',
  hideOnEscape = 1,
  hasEditBox = 1,
  OnAccept = function(self, data)
    amt = tonumber(self.editBox:GetText())
    if amt and amt > 0 then
      flsys:MassEP(amt, db.inprogress)
    else
      self.text = 'Please enter a number amount of EP to award',
      self:Show()
    end
  end,
  OnAlt = function(self, data)
    amt = tonumber(self.editBox:GetText())
    if amt and amt > 0 then
      flsys:StartRecur(0)
    else
      self.text = 'Please enter a number amount of EP to award',
      self:Show()
    end
  end
}

StaticPopupDialogs['DATARESET'] = {
  preferredIndex = STATICPOPUPS_NUMDIALOGS,
  text = 'Also remove all players from the database?',
  button1 = 'Yes',
  button2 = 'No',
  button3 = 'Cancel',
  timeout = 0,
  hideOnEscape = 1,
  whileDead = 1,
  OnAccept = function(self) flsys:ResetData(true) end,
  OnCancel = function(self) flsys:ResetData(false) end,
}

StaticPopupDialogs['LOGRESET'] = {
  preferredIndex = STATICPOPUPS_NUMDIALOGS,
  text = 'Delete all log info?',
  button1 = ACCEPT,
  button2 = CANCEL,
  timeout = 0,
  hideOnEscape = 1,
  whileDead = 1,
  OnAccept = function(self) flsys:ClearLog() end,
}

StaticPopupDialogs['RESET_GP'] = {
  preferredIndex = STATICPOPUPS_NUMDIALOGS,
  text = 'Reset all players\' GP to 0?  This cannot be undone.',
  button1 = ACCEPT,
  button2 = CANCEL,
  timeout = 0,
  hideOnEscape = 1,
  whileDead = 1,
  OnAccept = function(self) flsys:ResetGP() end,
}

StaticPopupDialogs['STARTRAID'] = {
  preferredIndex = STATICPOPUPS_NUMDIALOGS,
  text = 'Name of raid:',
  button1 = ACCEPT,
  button2 = CANCEL,
  timeout = 0,
  whileDead = 1,
  hideOnEscape = 1,
  hasEditBox = 1,
  maxLetters = 25,
  OnAccept = function(self)
    local text = self.editBox:GetText()
    flsys:StartRaid(flsys:Cap(text))
  end
}

StaticPopupDialogs['ENDRAID'] = {
  preferredIndex = STATICPOPUPS_NUMDIALOGS,
  text = 'End the raid?',
  button1 = ACCEPT,
  button2 = CANCEL,
  hideOnEscape = 1,
  timeout = 0,
  whileDead = 1,
  OnAccept = function(self)
    flsys:Info('Accepted ENDRAID dialog; ending')
    flsys:EndRaid()
  end,
  OnCancel = function()
    if not fld:IsEnabled() then
      flsys:ResumeRaid()
    end
  end,
}

StaticPopupDialogs['REMOVEPLAYER'] = {
  preferredIndex = STATICPOPUPS_NUMDIALOGS,
  text = 'Remove %s from the system? All of their data will be lost.',
  button1 = 'Yes',
  hideOnEscape = 1,
  button2 = 'No',
  timeout = 0,
  whileDead = 1,
  OnAccept = function(self)
    flsys:RemovePlayer(self.player)
  end,
}

StaticPopupDialogs['START_RECUR'] = {
  preferredIndex = STATICPOPUPS_NUMDIALOGS,
  text = 'Number of minutes until first award should go out:',
  button1 = ACCEPT,
  button2 = CANCEL,
  timeout = 0,
  whileDead = 1,
  enterClicksFirstButton = true,
  hasEditBox = 1,
  maxLetters = 2,
  OnShow = function(self, data)
    self.editBox:SetText('0')
  end,
  OnAccept = function(self, data)
    local minutes = tonumber(self.editBox:GetText())
    flsys:StartRecur(minutes)
  end
}

StaticPopupDialogs['ENDRECUR'] = {
  preferredIndex = STATICPOPUPS_NUMDIALOGS,
  text = 'End recurring EP award?',
  button1 = 'Yes',
  button2 = 'No',
  timeout = 0,
  enterClicksFirstButton = true,
  whileDead = 1,
  hideOnEscape = 1,
  OnAccept = function(self)
    flsys:EndRecur() end,
}

StaticPopupDialogs['RESUME_RAID'] = {
  preferredIndex = STATICPOPUPS_NUMDIALOGS,
  text = 'Resume your raid?',
  button1 = 'Yes',
  button2 = 'No',
  timeout = 0,
  whileDead = 1,
  OnAccept = function(self) flsys:ResumeRaid() end,
  OnCancel = function(self)
    flsys:Info('Cancelled RESUME_RAID prompt; ending raid')
    flsys:EndRaid()
  end,
}

StaticPopupDialogs['NOTDURINGRAID'] = {
  preferredIndex = STATICPOPUPS_NUMDIALOGS,
  text = 'You cannot do that while a raid is in progress.',
  button1 = 'Ok',
  timeout = 0,
  hideOnEscape = 1,
  enterClicksFirstButton = true,
  whileDead = 1,
}

StaticPopupDialogs['CONFIRM_RAIDSTOCOUNT'] = {
  preferredIndex = STATICPOPUPS_NUMDIALOGS,
  text = 'Change number of raids to count?',
  button1 = 'Yes',
  button2 = CANCEL,
  timeout = 0,
  enterClicksFirstButton = true,
  hideOnEscape = 1,
  whileDead = 1,
  OnAccept = function(self)
    flsys.db.profile.raidstocount = self.input
    flsys:GetConstants()
    end,
}

StaticPopupDialogs['DECAY'] = {
  preferredIndex = STATICPOPUPS_NUMDIALOGS,
  text = 'Decay ep and gp?',
  button1 = 'Yes',
  button2 = 'No',
  timeout = 0,
  hideOnEscape = 1,
  enterClicksFirstButton = true,
  whileDead = 1,
  OnAccept = function()
    flsys:DecayRaid()
  end,
}

StaticPopupDialogs['SENDDATA'] = {
  preferredIndex = STATICPOPUPS_NUMDIALOGS,
  text = 'Send data to:',
  button1 = ACCEPT,
  button2 = CANCEL,
  timeout = 0,
  whileDead = 1,
  hideOnEscape = 1,
  hasEditBox = 1,
  maxLetters = 25,
  OnAccept = function(self)
    local name = self.editBox:GetText()
    flsys:SendData(name)
  end
}

StaticPopupDialogs['RECEIVE_DATA'] = {
  preferredIndex = STATICPOPUPS_NUMDIALOGS,
  text = '%s is sending you loot standings. Accept them?\nWARNING: Imported data overwrites stored data!',
  button1 = 'Yes',
  button2 = 'No',
  timeout = 0,
  whileDead = 1,
  OnAccept = function(self)
    flsys:ImportEntered(self.text, self.player)
  end,
  OnCancel = function(self, reason)
    SendAddonMessage('flamingsys', 'datareply:'..'declined your data.', 'WHISPER', self.player)
  end,
}

StaticPopupDialogs['CHANGE_ILVL'] = {
  preferredIndex = STATICPOPUPS_NUMDIALOGS,
  text = '|cff6600FFflaming loot|r\n\nNew patch! Update your base ilvl to match current content',
  button1 = 'Okay',
  button2 = 'No thanks',
  OnAccept = function(self)
    flsys:ProcessChatCmd('opt')
  end,
  OnCancel = function(self, reason)
    flsys:Print('GP may become disproportiately high compared to EP if you don\'t adjust it.  Type |cffffff78/fs opt|r to edit your settings later.')
  end,
}
